# Copyright 2024 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import httpx
import json
import randomname

from typing import List, Union, Optional, Dict

from qiskit import qasm3, QuantumCircuit
from qiskit.result import Result
from qiskit.providers import JobV1
from qiskit.providers import JobError, JobTimeoutError, JobStatus

from qiskit_scaleway.versions import USER_AGENT

from scaleway_qaas_client.v1alpha1 import (
    QaaSClient,
    QaaSJobResult,
    QaaSJobData,
    QaaSJobClientData,
    QaaSCircuitData,
    QaaSJobRunData,
    QaaSJobBackendData,
    QaaSCircuitSerializationFormat,
)


class BaseJob(JobV1):
    def __init__(
        self,
        backend,
        client: QaaSClient,
        circuits: Union[List[QuantumCircuit], QuantumCircuit],
        config: Dict,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(backend, None)
        self._name = name or f"qj-qiskit-{randomname.get_name()}"
        self._client = client
        self._circuits = circuits
        self._config = config

    @property
    def name(self):
        return self._name

    def status(self) -> JobStatus:
        job = self._client.get_job(self._job_id)

        status_mapping = {
            "running": JobStatus.RUNNING,
            "waiting": JobStatus.QUEUED,
            "completed": JobStatus.DONE,
        }

        return status_mapping.get(job.status, JobStatus.ERROR)

    def submit(self, session_id: str) -> None:
        if self._job_id:
            raise RuntimeError(f"Job already submitted (ID: {self._job_id})")

        options = self._config.copy()

        run_data = QaaSJobRunData(
            options={
                "shots": options.pop("shots"),
                "memory": options.pop("memory", False),
            },
            circuits=list(
                map(
                    lambda c: QaaSCircuitData(
                        serialization_format=QaaSCircuitSerializationFormat.QASM_V3,
                        circuit_serialization=qasm3.dumps(c),
                    ),
                    self._circuits,
                )
            ),
        )

        backend_data = QaaSJobBackendData(
            name=self.backend().name,
            version=self.backend().version,
            options=options,
        )

        client_data = QaaSJobClientData(
            user_agent=USER_AGENT,
        )

        data = QaaSJobData.schema().dumps(
            QaaSJobData(
                backend=backend_data,
                run=run_data,
                client=client_data,
            )
        )

        model = self._client.create_model(
            payload=data,
        )

        if not model:
            raise RuntimeError("Failed to push circuit data")

        self._job_id = self._client.create_job(
            name=self._name,
            session_id=session_id,
            model_id=model.id,
        ).id

    def result(
        self, timeout: Optional[int] = None, fetch_interval: int = 3
    ) -> Union[Result, List[Result]]:
        if self._job_id == None:
            raise JobError("Job ID error")

        job_results = self._wait_for_result(timeout, fetch_interval)

        def __make_result_from_payload(payload: str) -> Result:
            payload_dict = json.loads(payload)

            return Result.from_dict(
                {
                    "results": payload_dict["results"],
                    "backend_name": self.backend().name,
                    "backend_version": self.backend().version,
                    "job_id": self._job_id,
                    "qobj_id": ", ".join(x.name for x in self._circuits),
                    "success": payload_dict["success"],
                    "header": payload_dict.get("header"),
                    "metadata": payload_dict.get("metadata"),
                }
            )

        qiskit_results = list(
            map(
                lambda r: __make_result_from_payload(
                    self._extract_payload_from_response(r)
                ),
                job_results,
            )
        )

        if len(qiskit_results) == 1:
            return qiskit_results[0]

        return qiskit_results

    def _extract_payload_from_response(self, job_result: QaaSJobResult) -> str:
        result = job_result.result

        if result is None or result == "":
            url = job_result.url

            if url is not None:
                resp = httpx.get(url.replace("http://s3", "http://localhost"))
                resp.raise_for_status()

                return resp.text
            else:
                raise Exception("Got result with empty data and url fields")
        else:
            return result

    def _wait_for_result(
        self, timeout: Optional[int], fetch_interval: int
    ) -> List[QaaSJobResult]:
        start_time = time.time()

        while True:
            elapsed = time.time() - start_time

            if timeout is not None and elapsed >= timeout:
                raise JobTimeoutError("Timed out waiting for result")

            status = self.status()

            if status == JobStatus.DONE:
                return self._client.list_job_results(self._job_id)

            if status == JobStatus.ERROR:
                raise JobError("Job error")

            time.sleep(fetch_interval)



# Copyright 2025 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import numpy as np
import random

from qiskit import QuantumCircuit
from qiskit_scaleway import ScalewayProvider


def _random_qiskit_circuit(size: int) -> QuantumCircuit:
    num_qubits = size
    num_gate = size

    qc = QuantumCircuit(num_qubits)

    for _ in range(num_gate):
        random_gate = np.random.choice(["unitary", "cx", "cy", "cz"])

        if random_gate == "cx" or random_gate == "cy" or random_gate == "cz":
            control_qubit = np.random.randint(0, num_qubits)
            target_qubit = np.random.randint(0, num_qubits)

            while target_qubit == control_qubit:
                target_qubit = np.random.randint(0, num_qubits)

            getattr(qc, random_gate)(control_qubit, target_qubit)
        else:
            for q in range(num_qubits):
                random_gate = np.random.choice(["h", "x", "y", "z"])
                getattr(qc, random_gate)(q)

    qc.measure_all()

    return qc


def test_aqt_multiple_circuits():
    provider = ScalewayProvider(
        project_id=os.environ["QISKIT_SCALEWAY_PROJECT_ID"],
        secret_key=os.environ["QISKIT_SCALEWAY_SECRET_KEY"],
        url=os.environ["QISKIT_SCALEWAY_API_URL"],
    )

    backend = provider.get_backend("aqt_ibex_simulation_pop_c16m128")

    assert backend is not None

    session_id = backend.start_session(
        name="my-aqt-session-autotest",
        deduplication_id=f"my-aqt-session-autotest-{random.randint(1, 1000)}",
        max_duration="15m",
    )

    assert session_id is not None

    try:
        qc1 = _random_qiskit_circuit(10)
        qc2 = _random_qiskit_circuit(12)
        qc3 = _random_qiskit_circuit(9)
        qc4 = _random_qiskit_circuit(10)

        run_result = backend.run(
            [qc1, qc2, qc3, qc4],
            shots=20,
            session_id=session_id,
        ).result()

        results = run_result.results

        assert len(results) == 4

        for result in results:
            assert result.success
    finally:
        backend.delete_session(session_id)



# Copyright 2024 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings

from typing import Union, Optional
from abc import ABC

from typing import Union, List

from qiskit.providers import BackendV2
from qiskit.circuit import QuantumCircuit

from scaleway_qaas_client.v1alpha1 import QaaSClient, QaaSPlatform

from .base_job import BaseJob


class BaseBackend(BackendV2, ABC):
    def __init__(
        self,
        provider,
        client: QaaSClient,
        platform: QaaSPlatform,
        **fields,
    ):
        super().__init__(
            provider=provider,
            backend_version=platform.version,
            name=platform.name,
            **fields,
        )

        self._platform = platform
        self._client = client

    @property
    def num_qubits(self) -> int:
        return self._platform.max_qubit_count

    @property
    def max_circuits(self) -> int:
        return self._platform.max_circuit_count

    @property
    def id(self) -> str:
        return self._platform.id

    @property
    def availability(self):
        return self._platform.availability

    def run(
        self, circuits: Union[QuantumCircuit, List[QuantumCircuit]], **run_options
    ) -> BaseJob:
        if not isinstance(circuits, List):
            circuits = [circuits]

        job_config = dict(self._options.items())

        for kwarg in run_options:
            if not hasattr(self.options, kwarg):
                warnings.warn(
                    f"Option {kwarg} is not used by this backend",
                    UserWarning,
                    stacklevel=2,
                )
            else:
                job_config[kwarg] = run_options[kwarg]

        session_id = job_config.get("session_id", None)

        job_config.pop("session_id")
        job_config.pop("session_name")
        job_config.pop("session_deduplication_id")
        job_config.pop("session_max_duration")
        job_config.pop("session_max_idle_duration")

        job = BaseJob(
            backend=self,
            client=self._client,
            circuits=circuits,
            config=job_config,
        )

        if session_id in ["auto", None]:
            session_id = self.start_session(name=f"auto-{self._options.session_name}")
            assert session_id is not None

        job.submit(session_id)

        return job

    def start_session(
        self,
        name: Optional[str] = None,
        deduplication_id: Optional[str] = None,
        max_duration: Union[int, str] = None,
        max_idle_duration: Union[int, str] = None,
    ) -> str:
        if name is None:
            name = self._options.get("session_name")

        if deduplication_id is None:
            deduplication_id = self._options.get("session_deduplication_id")

        if max_duration is None:
            max_duration = self._options.get("session_max_duration", "59m")

        if max_idle_duration is None:
            max_idle_duration = self._options.get("session_max_idle_duration", "59m")

        return self._client.create_session(
            name=name,
            platform_id=self.id,
            deduplication_id=deduplication_id,
            max_duration=max_duration,
            max_idle_duration=max_idle_duration,
        ).id

    def stop_session(self, session_id: str):
        self._client.terminate_session(
            session_id=session_id,
        )

    def delete_session(self, session_id: str):
        self._client.delete_session(
            session_id=session_id,
        )



from dataclasses import dataclass, field
from dataclasses_json import dataclass_json
from typing import List

from qiskit.transpiler import Target
from qiskit.circuit.library import get_standard_gate_name_mapping

from scaleway_qaas_client.v1alpha1 import QaaSPlatform


@dataclass_json
@dataclass
class _QiskitInstructionData:
    name: str
    params: List[str] = field(default=None)


@dataclass_json
@dataclass
class _QiskitTargetData:
    instructions: List[_QiskitInstructionData]


@dataclass_json
@dataclass
class _QiskitClientData:
    target: _QiskitTargetData


@dataclass_json
@dataclass
class _PlatformMetadata:
    qiskit: _QiskitClientData


def create_target_from_platform(platform: QaaSPlatform) -> Target:
    target = Target(num_qubits=platform.max_qubit_count)

    if not platform.metadata:
        return target

    metadata = _PlatformMetadata.from_json(platform.metadata)

    if not metadata:
        return target

    instructions = metadata.qiskit.target.instructions

    if not instructions:
        return target

    qiskit_gate_mapping = get_standard_gate_name_mapping()

    for instruction in instructions:
        qiskit_instruction = qiskit_gate_mapping.get(instruction.name)

        if not qiskit_instruction:
            raise Exception("could not find instruction:", instruction.name)

        target.add_instruction(qiskit_instruction)

    return target



# Copyright 2024 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib.metadata
import platform

from typing import Final

QISKIT_VERSION: Final = importlib.metadata.version("qiskit")
QISKIT_SCALEWAY_PROVIDER_VERSION: Final = importlib.metadata.version("qiskit-scaleway")

__version__: Final = QISKIT_SCALEWAY_PROVIDER_VERSION

USER_AGENT: Final = " ".join(
    [
        f"qiskit-scaleway/{QISKIT_SCALEWAY_PROVIDER_VERSION}",
        f"({platform.system()}; {platform.python_implementation()}/{platform.python_version()})",
        f"qiskit/{QISKIT_VERSION}",
    ]
)



# Copyright 2024 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from qiskit.primitives import BackendEstimatorV2
from qiskit.primitives.backend_estimator import (
    _prepare_counts,
    _run_circuits,
)


class Estimator(BackendEstimatorV2):
    def __init__(
        self,
        backend,
        session_id: str,
        options: dict | None = None,
    ):
        if not session_id:
            raise Exception("session_id must be not None")

        self._session_id = session_id

        super().__init__(
            backend=backend,
            options=options,
        )

    def _run_pubs(self, pubs, shots: int) -> list:
        """Compute results for pubs that all require the same value of ``shots``."""
        preprocessed_data = []
        flat_circuits = []
        for pub in pubs:
            data = self._preprocess_pub(pub)
            preprocessed_data.append(data)
            flat_circuits.extend(data.circuits)

        run_result, metadata = _run_circuits(
            flat_circuits,
            self._backend,
            shots=shots,
            seed_simulator=self._options.seed_simulator,
            session_id=self._session_id,
        )
        counts = _prepare_counts(run_result)

        results = []
        start = 0
        for pub, data in zip(pubs, preprocessed_data):
            end = start + len(data.circuits)
            expval_map = self._calc_expval_map(counts[start:end], metadata[start:end])
            start = end
            results.append(self._postprocess_pub(pub, expval_map, data, shots))

        return results



# Copyright 2025 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import numpy as np
import random

from qiskit import transpile
from qiskit.circuit.library import IQP
from qiskit.quantum_info import random_hermitian

from qiskit_scaleway import ScalewayProvider
from qiskit_scaleway.primitives import Sampler


def test_sampler():
    provider = ScalewayProvider(
        project_id=os.environ["QISKIT_SCALEWAY_PROJECT_ID"],
        secret_key=os.environ["QISKIT_SCALEWAY_SECRET_KEY"],
        url=os.environ["QISKIT_SCALEWAY_API_URL"],
    )

    backend = provider.get_backend("aer_simulation_pop_c16m128")

    assert backend is not None

    session_id = backend.start_session(
        name="my-sampler-session-autotest",
        deduplication_id=f"my-sampler-session-autotest-{random.randint(1, 1000)}",
        max_duration="15m",
    )

    assert session_id is not None

    try:
        sampler = Sampler(backend=backend, session_id=session_id)

        n_qubits = 10
        mat = np.real(random_hermitian(n_qubits, seed=1234))
        circuit = IQP(mat)
        circuit.measure_all()

        isa_circuit = transpile(circuit, backend=backend, optimization_level=1)
        job = sampler.run([isa_circuit], shots=100)
        result = job.result()

        assert result is not None
    finally:
        backend.stop_session(session_id)



# Copyright 2024 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from qiskit.primitives import BackendSamplerV2


class Sampler(BackendSamplerV2):
    def __init__(
        self,
        backend,
        session_id: str,
        options: dict | None = None,
    ):
        if not session_id:
            raise Exception("session_id must be not None")

        if not options:
            options = {}

        if not options.get("run_options"):
            options["run_options"] = {}

        if not options["run_options"].get("session_id"):
            options["run_options"]["session_id"] = session_id

        super().__init__(
            backend=backend,
            options=options,
        )



# Copyright 2024 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import randomname
import warnings

from typing import Union, List

from qiskit.providers import Options
from qiskit.circuit import QuantumCircuit
from qiskit.transpiler import Target

from qiskit_scaleway.backends.qsim.job import QsimJob
from qiskit_scaleway.backends import BaseBackend

from scaleway_qaas_client.v1alpha1 import QaaSClient, QaaSPlatform


class QsimBackend(BaseBackend):
    def __init__(self, provider, client: QaaSClient, platform: QaaSPlatform):
        super().__init__(
            provider=provider,
            client=client,
            platform=platform,
        )

        self._options = self._default_options()
        self.options.set_validator("shots", (1, platform.max_shot_count))

        self._target = Target(num_qubits=platform.max_qubit_count)

    def __repr__(self) -> str:
        return f"<QsimBackend(name={self.name},num_qubits={self.num_qubits},platform_id={self.id})>"

    @property
    def target(self):
        return self._target

    @property
    def job_cls(self):
        return QsimJob

    def run(
        self, circuits: Union[QuantumCircuit, List[QuantumCircuit]], **kwargs
    ) -> QsimJob:
        if not isinstance(circuits, List):
            circuits = [circuits]

        job_config = dict(self._options.items())

        for kwarg in kwargs:
            if not hasattr(self.options, kwarg):
                warnings.warn(
                    f"Option {kwarg} is not used by this backend",
                    UserWarning,
                    stacklevel=2,
                )
            else:
                job_config[kwarg] = kwargs[kwarg]

        job_name = f"qj-qsim-{randomname.get_name()}"

        session_id = job_config.get("session_id", None)

        job_config.pop("session_id")
        job_config.pop("session_name")
        job_config.pop("session_deduplication_id")
        job_config.pop("session_max_duration")
        job_config.pop("session_max_idle_duration")

        job = QsimJob(
            backend=self,
            client=self._client,
            circuits=circuits,
            config=job_config,
            name=job_name,
        )

        if session_id in ["auto", None]:
            session_id = self.start_session(name=f"auto-{self._options.session_name}")
            assert session_id is not None

        job.submit(session_id)

        return job

    @classmethod
    def _default_options(self):
        return Options(
            session_id="auto",
            session_name="qs-qiskit-qsim",
            session_deduplication_id="qs-qiskit-qsim",
            session_max_duration="59m",
            session_max_idle_duration="59m",
            shots=1000,
            circuit_memoization_size=0,
            max_fused_gate_size=2,
            ev_noisy_repetitions=1,
            denormals_are_zeros=False,
        )



# Copyright 2025 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from qiskit.providers import Options

from qiskit_scaleway.backends import BaseBackend
from qiskit_scaleway.utils import create_target_from_platform

from scaleway_qaas_client.v1alpha1 import QaaSClient, QaaSPlatform


class IqmBackend(BaseBackend):
    def __init__(self, provider, client: QaaSClient, platform: QaaSPlatform):
        super().__init__(
            provider=provider,
            client=client,
            platform=platform,
        )

        self._options = self._default_options()

        self._platform = platform
        self._target = create_target_from_platform(self._platform)

        self._options.max_shots = platform.max_shot_count
        self._options.set_validator("shots", (1, platform.max_shot_count))

    def __repr__(self) -> str:
        return f"<IqmBackend(name={self.name},num_qubits={self.num_qubits},platform_id={self.id})>"

    @property
    def target(self):
        return self._target

    @classmethod
    def _default_options(self):
        return Options(
            session_id="auto",
            session_name="qs-qiskit-iqm",
            session_deduplication_id="qs-qiskit-iqm",
            session_max_duration="59h",
            session_max_idle_duration="59m",
            shots=1000,
        )



# Copyright 2024 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from qiskit.providers import Options, convert_to_target

from qiskit_aer.backends.aer_simulator import BASIS_GATES, AerBackendConfiguration
from qiskit_aer.backends.aerbackend import NAME_MAPPING

from qiskit_scaleway.backends import BaseBackend

from scaleway_qaas_client.v1alpha1 import QaaSClient, QaaSPlatform


class AerBackend(BaseBackend):
    def __init__(self, provider, client: QaaSClient, platform: QaaSPlatform):
        super().__init__(
            provider=provider,
            client=client,
            platform=platform,
        )

        self._options = self._default_options()
        self._configuration = AerBackendConfiguration.from_dict(
            {
                "open_pulse": False,
                "backend_name": platform.name,
                "backend_version": platform.version,
                "n_qubits": platform.max_qubit_count,
                "url": "https://github.com/Qiskit/qiskit-aer",
                "simulator": True,
                "local": False,
                "conditional": True,
                "memory": True,
                "max_shots": platform.max_shot_count,
                "description": platform.description,
                "coupling_map": None,
                "basis_gates": BASIS_GATES["automatic"],
                "gates": [],
            }
        )
        self._properties = None
        self._target = convert_to_target(
            self._configuration, self._properties, None, NAME_MAPPING
        )

    def __repr__(self) -> str:
        return f"<AerBackend(name={self.name},num_qubits={self.num_qubits},platform_id={self.id})>"

    @property
    def target(self):
        return self._target

    @classmethod
    def _default_options(self):
        return Options(
            session_id="auto",
            session_name="qs-qiskit-aer",
            session_deduplication_id="qs-qiskit-aer",
            session_max_duration="59m",
            session_max_idle_duration="59m",
            shots=1000,
            memory=False,
            seed_simulator=None,
            method="automatic",
            precision="double",
            max_shot_size=None,
            enable_truncation=True,
            max_parallel_experiments=1,
            zero_threshold=1e-10,
            validation_threshold=1e-8,
            accept_distributed_results=None,
            runtime_parameter_bind_enable=False,
            statevector_parallel_threshold=14,
            statevector_sample_measure_opt=10,
            stabilizer_max_snapshot_probabilities=32,
            extended_stabilizer_sampling_method="resampled_metropolis",
            extended_stabilizer_metropolis_mixing_time=5000,
            extended_stabilizer_approximation_error=0.05,
            extended_stabilizer_norm_estimation_samples=100,
            extended_stabilizer_norm_estimation_repetitions=3,
            extended_stabilizer_parallel_threshold=100,
            extended_stabilizer_probabilities_snapshot_samples=3000,
            matrix_product_state_max_bond_dimension=None,
            matrix_product_state_truncation_threshold=1e-16,
            mps_sample_measure_algorithm="mps_apply_measure",
            mps_log_data=False,
            mps_swap_direction="mps_swap_left",
            chop_threshold=1e-8,
            mps_parallel_threshold=14,
            mps_omp_threads=1,
            tensor_network_num_sampling_qubits=10,
            use_cuTensorNet_autotuning=False,
            fusion_enable=True,
            fusion_verbose=False,
            fusion_max_qubit=None,
            fusion_threshold=None,
        )



# Copyright 2025 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import random

from scipy.optimize import minimize

from qiskit import QuantumCircuit
from qiskit.quantum_info.operators.base_operator import BaseOperator
from qiskit.quantum_info import SparsePauliOp
from qiskit.circuit.library import TwoLocal

from qiskit_scaleway import ScalewayProvider
from qiskit_scaleway.primitives import Estimator


def test_estimator():
    provider = ScalewayProvider(
        project_id=os.environ["QISKIT_SCALEWAY_PROJECT_ID"],
        secret_key=os.environ["QISKIT_SCALEWAY_SECRET_KEY"],
        url=os.environ["QISKIT_SCALEWAY_API_URL"],
    )

    backend = provider.get_backend("aer_simulation_pop_c16m128")

    assert backend is not None

    session_id = backend.start_session(
        name="my-estimator-session-autotest",
        deduplication_id=f"my-estimator-session-autotest-{random.randint(1, 1000)}",
        max_duration="15m",
    )

    assert session_id is not None

    try:
        # Buildup the estimator for the VQE
        estimator = Estimator(backend=backend, session_id=session_id)

        # Specify the problem Hamiltonian
        hamiltonian = SparsePauliOp.from_list(
            [
                ("II", -1.052373245772859),
                ("IZ", 0.39793742484318045),
                ("ZI", -0.39793742484318045),
                ("ZZ", -0.01128010425623538),
                ("XX", 0.18093119978423156),
            ]
        )

        # Define the VQE Ansatz, initial point, and cost function
        ansatz = TwoLocal(num_qubits=2, rotation_blocks="ry", entanglement_blocks="cz")
        initial_point = [0] * 8

        def cost_function(
            params,
            ansatz: QuantumCircuit,
            hamiltonian: BaseOperator,
            estimator: Estimator,
        ) -> float:
            """Cost function for the VQE.

            Return the estimated expectation value of the Hamiltonian
            on the state prepared by the Ansatz circuit.
            """
            result = estimator.run([(ansatz, hamiltonian, params)]).result()

            loss = float(result[0].data["evs"])

            print(f"> loss {loss}")

            return loss

        # Run the VQE using the SciPy minimizer routine
        result = minimize(
            cost_function,
            initial_point,
            args=(ansatz, hamiltonian, estimator),
            method="cobyla",
        )

        assert result is not None
        assert result.fun is not None

        # Print the found minimum eigenvalue
        print(f"> {result.fun}")
    finally:
        backend.stop_session(session_id)



# Copyright 2025 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import random

from qiskit import QuantumCircuit
from qiskit_scaleway import ScalewayProvider


def test_qsim_simple_circuit():
    provider = ScalewayProvider(
        project_id=os.environ["QISKIT_SCALEWAY_PROJECT_ID"],
        secret_key=os.environ["QISKIT_SCALEWAY_SECRET_KEY"],
        url=os.environ["QISKIT_SCALEWAY_API_URL"],
    )

    backend = provider.get_backend("qsim_simulation_pop_c16m128")

    assert backend is not None

    session_id = backend.start_session(
        name="my-qsim-session-autotest",
        deduplication_id=f"my-qsim-session-autotest-{random.randint(1, 1000)}",
        max_duration="15m",
    )

    assert session_id is not None

    try:
        qc = QuantumCircuit(4)
        qc.h(0)
        qc.cx(0, 1)
        qc.cx(0, 2)
        qc.cx(0, 3)
        qc.measure_all()

        shots_count = 1000
        job = backend.run(qc, shots=shots_count, session_id=session_id)

        # cirq_result = job.result(format="cirq")

        # assert cirq_result is not None
        # assert cirq_result.repetitions == shots_count

        qiskit_result = job.result(format="qiskit")

        assert qiskit_result is not None
        assert qiskit_result.success
        assert qiskit_result.results[0].shots == shots_count

    finally:
        backend.delete_session(session_id)



# Copyright 2024 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from typing import Optional, List, Dict

from qiskit.providers.providerutils import filter_backends

from qiskit_scaleway.backends import (
    BaseBackend,
    IqmBackend,
    AqtBackend,
    QsimBackend,
    AerBackend,
)

from scaleway_qaas_client.v1alpha1 import QaaSClient


class ScalewayProvider:
    """
    :param project_id: optional UUID of the Scaleway Project, if the provided ``project_id`` is None, the value is loaded from the QISKIT_SCALEWAY_PROJECT_ID environment variables

    :param secret_key: optional authentication token required to access the Scaleway API, if the provided ``secret_key`` is None, the value is loaded from the QISKIT_SCALEWAY_SECRET_KEY environment variables

    :param url: optional value, endpoint URL of the API, if the provided ``url`` is None, the value is loaded from the QISKIT_SCALEWAY_API_URL environment variables
    """

    def __init__(
        self,
        project_id: Optional[str] = None,
        secret_key: Optional[str] = None,
        url: Optional[str] = None,
    ) -> None:
        secret_key = secret_key or os.getenv("QISKIT_SCALEWAY_SECRET_KEY")
        project_id = project_id or os.getenv("QISKIT_SCALEWAY_PROJECT_ID")
        url = url or os.getenv("QISKIT_SCALEWAY_API_URL")

        if secret_key is None:
            raise Exception("secret_key is missing")

        if project_id is None:
            raise Exception("project_id is missing")

        self.__client = QaaSClient(
            url=url, secret_key=secret_key, project_id=project_id
        )

    def get_backend(self, name=None, **kwargs):
        """Return a single backend matching the specified filtering.

        Args:
            name (str): name of the backend.
            **kwargs: dict used for filtering.

        Returns:
            Backend: a backend matching the filtering.

        Raises:
            Exception: if no backend could be found or
                more than one backend matches the filtering criteria.
        """
        backends = self.backends(name, **kwargs)
        if len(backends) > 1:
            raise Exception("More than one backend matches the criteria")
        if not backends:
            raise Exception("No backend matches the criteria")

        return backends[0]

    def backends(self, name: Optional[str] = None, **kwargs) -> List[BaseBackend]:
        """Return a list of backends matching the specified filtering.

        Args:
            name (str): name of the backend.

        Returns:
            list[ScalewayBackend]: a list of Backends that match the filtering
                criteria.
        """

        scaleway_backends = []
        filters = {}

        if kwargs.get("operational") is not None:
            filters["operational"] = kwargs.pop("operational", None)

        if kwargs.get("min_num_qubits") is not None:
            filters["min_num_qubits"] = kwargs.pop("min_num_qubits", None)

        platforms = self.__client.list_platforms(name=name)

        for platform in platforms:
            if platform.backend_name == "aer":
                scaleway_backends.append(
                    AerBackend(
                        provider=self,
                        client=self.__client,
                        platform=platform,
                    )
                )
            elif platform.backend_name == "qsim":
                scaleway_backends.append(
                    QsimBackend(
                        provider=self,
                        client=self.__client,
                        platform=platform,
                    )
                )
            elif platform.provider_name == "aqt":
                scaleway_backends.append(
                    AqtBackend(
                        provider=self,
                        client=self.__client,
                        platform=platform,
                    )
                )
            elif platform.provider_name == "iqm":
                scaleway_backends.append(
                    IqmBackend(
                        provider=self,
                        client=self.__client,
                        platform=platform,
                    )
                )

        if filters is not None:
            scaleway_backends = self.filters(scaleway_backends, filters)

        return filter_backends(scaleway_backends, **kwargs)

    def filters(self, backends: List[BaseBackend], filters: Dict) -> List[BaseBackend]:
        operational = filters.get("operational")
        min_num_qubits = filters.get("min_num_qubits")

        if operational is not None:
            backends = [
                b for b in backends if b.availability in ["available", "scarce"]
            ]

        if min_num_qubits is not None:
            backends = [b for b in backends if b.num_qubits >= min_num_qubits]

        return backends



# Copyright 2025 Scaleway
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from qiskit.primitives import BackendEstimator


class Estimator(BackendEstimator):
    def __init__(
        self,
        backend,
        session_id: str,
        options: dict | None = None,
        abelian_grouping: bool = True,
        skip_transpilation: bool = False,
    ):
        if not session_id:
            raise Exception("session_id must be not None")

        self._session_id = session_id

        super().__init__(
            backend,
            bound_pass_manager=backend.get_pass_manager(),
            options=options,
            abelian_grouping=abelian_grouping,
            skip_transpilation=skip_transpilation,
        )

    def _run(self, circuits, observables, parameter_values, **run_options):
        run_options["session_id"] = self._session_id
        return super()._run(circuits, observables, parameter_values, **run_options)
